﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("DuplicateSolutionsAnalyzer", "An operator that analyzes duplicate solutions in the population.")]
  [StorableClass]
  public class DuplicateSolutionsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return false; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public IValueParameter<ISingleObjectiveSolutionSimilarityCalculator> SimilarityCalculatorParameter {
      get { return (IValueParameter<ISingleObjectiveSolutionSimilarityCalculator>)Parameters["SimilarityCalculator"]; }
    }
    public ILookupParameter<ItemCollection<IItem>> OperatorsParameter {
      get { return (ILookupParameter<ItemCollection<IItem>>)Parameters["Operators"]; }
    }
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters["UpdateInterval"]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["Generations"]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }

    [Storable]
    private DataTableHelper chartingHelper, similarityChartHelper;
    #endregion

    [StorableConstructor]
    private DuplicateSolutionsAnalyzer(bool deserializing) : base(deserializing) { }

    private DuplicateSolutionsAnalyzer(DuplicateSolutionsAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      chartingHelper = (DataTableHelper)original.chartingHelper.Clone(cloner);
      similarityChartHelper = (DataTableHelper)original.similarityChartHelper.Clone(cloner);
    }

    public DuplicateSolutionsAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new ValueParameter<ISingleObjectiveSolutionSimilarityCalculator>("SimilarityCalculator"));
      Parameters.Add(new LookupParameter<ItemCollection<IItem>>("Operators", "The operators and items that the problem provides to the algorithms."));
      Parameters.Add(new ValueParameter<IntValue>("UpdateInterval", "The interval in which the operator should be applied.", new IntValue(2)));
      Parameters.Add(new LookupParameter<IntValue>("Generations", "Nr of generations."));

      chartingHelper = new DataTableHelper();
      similarityChartHelper = new DataTableHelper();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DuplicateSolutionsAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      if (SimilarityCalculatorParameter.Value == null) {
        SimilarityCalculatorParameter.Value = OperatorsParameter.ActualValue.OfType<ISingleObjectiveSolutionSimilarityCalculator>().FirstOrDefault();
      }
      chartingHelper.InitializeChart(Results, "Duplicate Solutions", new string[] { "Duplicate Solutions per Generation" });
      similarityChartHelper.InitializeChart(Results, "Average Similarity", new string[] { "Avg. Similarity per Generation" });

      if (GenerationsParameter.ActualValue.Value % UpdateIntervalParameter.Value.Value == 0) {
        var similarities = SimilarityCalculatorParameter.Value.CalculateSolutionCrowdSimilarity(ExecutionContext.Scope);
        List<double> similaritiesArr = new List<double>();
        int counter = 0;
        for (int i = 0; i < similarities.Length; i++) {
          for (int j = 0; j < similarities[i].Length; j++) {
            if (similarities[i][j] == 1.0 && i != j) {
              counter++;
              break;
            }
          }
          for (int j = 0; j < similarities[i].Length; j++) {
            if (i != j) {
              similaritiesArr.Add(similarities[i][j]);
            }
          }
        }

        similarityChartHelper.AddPoint(similaritiesArr.Average());
        chartingHelper.AddPoint(counter / (double)similarities.Length);
      }
      return base.Apply();
    }
  }
}
